@@ -6,9 +6,9 @@
 #![feature(const_raw_ptr_deref)]
 
 use components_arena::{Arena, Component, NewtypeComponentId, Id};
-use dep_obj::{DepObjBaseBuilder, DepObjId, Items, dep_obj, dep_type, dep_type_with_builder, Change};
+use dep_obj::{Change, DepObjBaseBuilder, DepObjId, dep_obj, dep_type, dep_type_with_builder, ItemChange};
 use macro_attr_2018::macro_attr;
-use dep_obj::binding::{Bindings, EventBinding0, EventBinding1};
+use dep_obj::binding::{Binding1, Binding2, BindingExt2, Bindings, b_yield, b_immediate};
 use dyn_context::state::{State, StateExt};
 use std::any::{TypeId, Any};
 use std::borrow::Cow;
@@ -116,46 +116,30 @@ impl Npc {
     fn new(state: &mut dyn State) -> Npc {
         let game: &mut Game = state.get_mut();
         let npc = game.npcs.insert(|id| (NpcComponent { props: NpcProps::new_priv() }, Npc(id)));
-        let removed_items_binding = EventBinding0::new(state, (), |state, (), items: Option<&mut Items<Item>>| {
-            items.map(|items| {
-                for item in items.iter() {
-                    ItemProps::EQUIPPED.set(state, item.props(), false);
-                }
-            })
-        });
-        removed_items_binding.set_event_source(state, &mut NpcProps::EQUIPPED_ITEMS.removed_items_source(npc.props()));
-        npc.props().add_binding(state, removed_items_binding);
-        let inserted_items_binding = EventBinding0::new(state, (), |state, (), items: Option<&mut Items<Item>>| {
-            items.map(|items| {
-                for item in items.iter() {
-                    ItemProps::EQUIPPED.set(state, item.props(), true);
-                }
-            })
-        });
-        inserted_items_binding.set_event_source(state, &mut NpcProps::EQUIPPED_ITEMS.inserted_items_source(npc.props()));
-        npc.props().add_binding(state, inserted_items_binding);
-        let apply_enhancement = EventBinding1::new(state, npc, |state, npc, enhancement, items: Option<&mut Items<Item>>| {
-            if let Some(items) = items {
-                for item in items.iter() {
-                    ItemProps::ENHANCEMENT.set(state, item.props(), enhancement);
+
+        let equipped = Binding1::new(state, (), |(), change: Option<ItemChange<Item>>|
+            change.filter(|change| !change.is_update())
+        );
+        equipped.dispatch(state, (), |state, (), change|
+            ItemProps::EQUIPPED.set(state, change.item.props(), change.is_insert())
+        );
+        equipped.set_source_1(state, &mut NpcProps::EQUIPPED_ITEMS.item_source(npc.props()));
+        npc.props().add_binding(state, equipped);
+
+        let enhancement = BindingExt2::new(state, (), |state, (), enhancement, change: Option<ItemChange<Item>>| {
+            if let Some(change) = change {
+                if change.is_remove() {
+                    ItemProps::ENHANCEMENT.unset(state, change.item.props())
+                } else {
+                    ItemProps::ENHANCEMENT.set(state, change.item.props(), enhancement)
                 }
             } else {
-                NpcProps::EQUIPPED_ITEMS.refresh(state, npc.props());
+                b_yield(())
             }
-            Some(())
         });
-        apply_enhancement.set_source_1(state, &mut NpcProps::ITEMS_ENHANCEMENT.value_source(npc.props()));
-        apply_enhancement.set_event_source(state, &mut NpcProps::EQUIPPED_ITEMS.inserted_items_source(npc.props()));
-        npc.props().add_binding(state, apply_enhancement);
-        let unapply_enhancement = EventBinding0::new(state, (), |state, (), items: Option<&mut Items<Item>>| {
-            items.map(|items| {
-                for item in items.iter() {
-                    ItemProps::ENHANCEMENT.unset(state, item.props());
-                }
-            })
-        });
-        unapply_enhancement.set_event_source(state, &mut NpcProps::EQUIPPED_ITEMS.removed_items_source(npc.props()));
-        npc.props().add_binding(state, unapply_enhancement);
+        enhancement.set_source_2(state, &mut NpcProps::EQUIPPED_ITEMS.item_source_with_update(enhancement, npc.props()));
+        enhancement.set_source_1(state, &mut NpcProps::ITEMS_ENHANCEMENT.value_source(npc.props()));
+
         npc
     }
 
@@ -220,6 +204,7 @@ impl Game {
 fn main() {
     let game = &mut Game::new();
     let npc = Npc::new(game);
+    b_immediate(NpcProps::ITEMS_ENHANCEMENT.set(game, npc.props(), 5));
     let sword = Item::new(game);
     sword.build(game, |sword| sword
         .props(|props| props
@@ -227,28 +212,47 @@ fn main() {
         )
     );
     let shield = Item::new(game);
-    ItemProps::NAME.set(game, shield.props(), Cow::Borrowed("Shield"));
+    b_immediate(ItemProps::NAME.set(game, shield.props(), Cow::Borrowed("Shield")));
     for item in [sword, shield] {
-        let log = EventBinding1::new(game, (), |_, (), name, equipped: Option<&mut Change<bool>>|
-            equipped.map(|equipped| (*equipped.new(), name))
+        let equipped = Binding2::new(game, (), |(), name, equipped: Option<Change<bool>>|
+            equipped.map(|equipped| (name, equipped.new))
         );
-        log.set_event_source(game, &mut ItemProps::EQUIPPED.change_source(item.props()));
-        log.set_source_1(game, &mut ItemProps::NAME.value_source(item.props()));
-        log.set_target_fn(game, (), |game, (), (equipped, name)| {
+        equipped.set_target_fn(game, (), |game, (), (name, equipped)| {
             let game: &mut Game = game.get_mut();
             writeln!(&mut game.log, "{} {}.", name, if equipped { "equipped" } else { "unequipped" }).unwrap();
         });
-        item.props().add_binding(game, log);
+        equipped.set_source_1(game, &mut ItemProps::NAME.value_source(item.props()));
+        equipped.set_source_2(game, &mut ItemProps::EQUIPPED.change_source(item.props()));
+        item.props().add_binding(game, equipped);
+        let enhancement = Binding2::new(game, (), |(), name, enhancement: Option<Change<i8>>|
+            enhancement.map(|enhancement| (name, enhancement))
+        );
+        enhancement.set_target_fn(game, (), |game, (), (name, enhancement)| {
+            let game: &mut Game = game.get_mut();
+            writeln!(&mut game.log, "{} enhancement changed: {} -> {}.", name, enhancement.old, enhancement.new).unwrap();
+        });
+        enhancement.set_source_1(game, &mut ItemProps::NAME.value_source(item.props()));
+        enhancement.set_source_2(game, &mut ItemProps::ENHANCEMENT.change_source(item.props()));
+        item.props().add_binding(game, enhancement);
     }
-    NpcProps::EQUIPPED_ITEMS.push(game, npc.props(), sword);
-    NpcProps::EQUIPPED_ITEMS.push(game, npc.props(), shield);
-    NpcProps::EQUIPPED_ITEMS.remove(game, npc.props(), 0);
-    assert_eq!(game.log, "\
-        Sword equipped.\n\
-        Shield equipped.\n\
-        Sword unequipped.\n\
-    ");
+    b_immediate(NpcProps::EQUIPPED_ITEMS.push(game, npc.props(), sword));
+    b_immediate(NpcProps::EQUIPPED_ITEMS.push(game, npc.props(), shield));
+    b_immediate(NpcProps::ITEMS_ENHANCEMENT.set(game, npc.props(), 4));
+    b_immediate(NpcProps::EQUIPPED_ITEMS.remove(game, npc.props(), 0));
+    b_immediate(NpcProps::ITEMS_ENHANCEMENT.set(game, npc.props(), 5));
     npc.drop_npc(game);
     sword.drop_item(game);
     shield.drop_item(game);
+    print!("{}", game.log);
+    assert_eq!(game.log, "\
+        Sword equipped.\n\
+        Sword enhancement changed: 0 -> 5.\n\
+        Shield equipped.\n\
+        Shield enhancement changed: 0 -> 5.\n\
+        Sword enhancement changed: 5 -> 4.\n\
+        Shield enhancement changed: 5 -> 4.\n\
+        Sword unequipped.\n\
+        Sword enhancement changed: 4 -> 0.\n\
+        Shield enhancement changed: 4 -> 5.\n\
+    ");
 }
